import os
import mutagen
import mutagen.id3
from logbook import warn


FIELDS = ('title',
          'artist',
          'album',
          'genre',
          'albumartist',
          'bpm',
          'key',
          'tracknumber',
          'year')


class NoTaggerError(Exception):
    pass


class Tagger(object):

    def __init__(self, fname):
        self.tags = mutagen.File(fname)

    def save(self):
        self.tags.save()

    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __str__(self):
        return str(self.todict())

    def todict(self):
        return {key: self[key] for key in FIELDS}


class Mp3Tagger(Tagger):

    supported_extensions = ('.mp3',)

    def __getitem__(self, key):
        try:
            return self.tags[self.translate_key(key)].text[0]
        except KeyError:
            # This should not happen for correcly formatted tags, but it seems
            # to happen for some files
            return None

    def __setitem__(self, key, value):
        tagname = self.translate_key(key)
        self.tags[tagname] = getattr(mutagen.id3, tagname)(text=[value])

    def translate_key(self, key):
        if self.tags.tags is None:
            year_key = 'TDRC'
        elif self.tags.tags.version == (2, 4, 0):
            year_key = 'TDRC'
        else:
            year_key = 'TYER'
        d = {'title': 'TIT2',
             'artist': 'TPE1',
             'album': 'TALB',
             'genre': 'TCON',
             'albumartist': 'TPE2',
             'bpm': 'TBPM',
             'tracknumber': 'TRCK',
             'year': year_key,
             'key': 'TKEY',
             }
        return d[key]


class FlacTagger(Tagger):

    supported_extensions = ('.flac', '.ogg')

    def __getitem__(self, key):
        if key in self.tags:
            return self.tags[key][0]
        else:
            return None

    def __setitem__(self, key, value):
        self.tags[key] = [value]


class M4aTagger(Tagger):

    supported_extensions = ('.m4a',)

    def __getitem__(self, key):
        if key == 'tracknumber':
            return str(self.tags['trkn'][0][0])
        elif key == 'key':
            warn('Keys are not supported for M4A files')
            return ''
        return self.tags[self.translate_key(key)][0]

    def __setitem__(self, key, value):
        if key == 'tracknumber':
            k, n = self.tags['trkn'][0]
            self.tags['trkn'][0] = (int(value), n)
        elif key == 'key':
            warn('Keys are not supported for M4A files')
            pass
        else:
            self.tags[self.translate_key(key)] = [value]

    def translate_key(self, key):
        self.d = {'title': '\xa9nam',
                  'artist': '\xa9ART',
                  'album': '\xa9alb',
                  'genre': '\xa9gen',
                  'albumartist': 'aART',
                  'bpm': 'tmpo',
                  'year': '\xa9day',
                  }
        if key == 'key' and key not in self.d:
            candidates = [key for key in self.tags if 'initialkey' in key]
            self.d['key'] = candidates[0]
        return self.d[key]


def get_tagger(fname):
    basename, extension = os.path.splitext(fname)
    tagger = [tagr
              for tagr in Tagger.__subclasses__()
              if extension.lower() in tagr.supported_extensions]
    if len(tagger):
        return tagger[0](fname)
    else:
        raise NoTaggerError('Could not find tagger for extension {}'
                            .format(extension))


def bpm2str(value):
    return str(int(round(float(value))))


def set_multiple_tags(fname, tags, prefix=''):
    tagger = get_tagger(fname)
    changed = False
    for key in FIELDS:
        value = tags.get(prefix + key, None)
        if value is not None:
            tagger[key] = value
            changed = True
    if changed:
        tagger.save()



import os
import sys
import yaml
from contextlib import contextmanager
from logbook import info, warning

from usiq import tagger, parser, renamer


class UsiqError(Exception):
    pass


def show(fname):
    info(tagger.get_tagger(fname))


def tag(fnames, args):
    default_tags = {key: args['--' + key]
                    for key in tagger.FIELDS
                    if '--' + key in args and args['--' + key] is not None}

    if args['--import']:
        with open_file_or_stdinout(args['--import']) as f:
            filetags = yaml.load(f)
    else:
        filetags = {}

    for fname in fnames:
        abs_fname = os.path.abspath(fname)
        tags = filetags[abs_fname] if abs_fname in filetags else {}
        tags.update(default_tags.copy())
        if args['--pattern']:
            tags.update(parser.parse_filename(fname, args['--pattern']))

        info('Setting tags {} on file {}'.format(tags, fname))
        if not args['--dry']:
            tagger.set_multiple_tags(fname, tags, prefix='')


def rename(fnames, args):
    pattern = args['--pattern']
    if illegal_pattern(pattern):
        raise UsiqError('Illegal pattern, aborting')

    for fname in fnames:
        tags = tagger.get_tagger(fname)
        new_fname = renamer.create_filename(tags, pattern)
        _, extension = os.path.splitext(fname)
        new_fname += extension
        target_exists = os.path.exists(new_fname)
        if target_exists:
            warning('Not moving {} -> {}, target file exists!'
                    .format(fname, new_fname))
            continue
        info('Moving {} -> {}'.format(fname, new_fname))
        if not args['--dry']:
            os.makedirs(os.path.dirname(new_fname), exist_ok=True)
            os.rename(fname, new_fname)


def export(fnames, args):
    outfile = args['--output']
    with open_file_or_stdinout(outfile, 'w') as f:
        yaml.dump({os.path.abspath(fname): tagger.get_tagger(fname).todict()
                   for fname in fnames},
                  f,
                  default_flow_style=False)


def illegal_pattern(pattern):
    constant = len(parser.get_fields(pattern)) == 0
    _, extension = os.path.splitext(pattern)
    has_extension = extension.lower() in ['.mp3', '.flac', '.ogg', '.m4a']
    return constant or has_extension


def with_config(args):
    if os.path.exists(args['--config']):
        with open(args['--config']) as f:
            cfg = yaml.load(f)
    else:
        cfg = {}
    for key, value in args.items():
        if key not in cfg:
            cfg[key] = value
        elif value is not None:
            cfg[key] = value
    return cfg


@contextmanager
def open_file_or_stdinout(fname, mode='r', **kwargs):
    if fname == '-':
        if 'r' in mode:
            yield sys.stdin
        elif 'w' in mode or 'a' in mode:
            yield sys.stdout
        else:
            yield sys.stdin
    else:
        f = open(fname, mode, **kwargs)
        yield f
        f.close()



import os
import string
from . import parser


def create_filename(tags, pattern):
    fields = parser.get_fields(pattern)
    for field, formatter in fields.items():
        if formatter:
            pattern = pattern.replace(
                '<{}.{}>'.format(field, formatter),
                format_filename(getattr(tags[field], formatter)()))
        else:
            pattern = pattern.replace(
                '<{}>'.format(field),
                format_filename(tags[field]))
    return os.path.expanduser(pattern)


def format_filename(filename):
    filename = filename.replace('[', '(').replace(']', ')')
    filename = format_quotes(filename)
    valid_chars = '-_.() {}{}'.format(string.ascii_letters, string.digits)
    filename = ''.join(c if c in valid_chars else '_' for c in filename)
    return filename


def format_quotes(filename):
    if '"' not in filename:
        return filename

    if len([c for c in filename if c == '"']) == 1:
        return filename.replace('"', 'in')

    output = []
    last_char = ''
    open_quote = False
    for i, c in enumerate(filename):
        if c == '"':
            if open_quote:
                # Assume that we definitely want to close any quote first
                open_quote = False
            elif last_char == ' ':
                # This is almost definitely an opening quote
                open_quote = True
            elif last_char.isdigit():
                # Almost sure to be inches
                output.append('in')
        else:
            output.append(c)
        last_char = c
    return ''.join(output)



import re
import os

from .tagger import FIELDS


def parse_filename(fname, pattern):
    basename, _ = os.path.splitext(os.path.abspath(fname))
    regexp = construct_regexp(pattern)
    parsed = re.search(regexp, basename).groupdict()
    for key in parsed:
        parsed[key] = parsed[key].replace('_', ' ')
    return parsed


def construct_regexp(pattern):
    fields = set(get_fields(pattern))
    regexp = pattern.replace('(', r'\(').replace(')', r'\)')
    regexp = regexp.replace('<__any__>', '[^/]+')
    number_fields = {'bpm', 'year', 'tracknumber'}
    complex_fields = number_fields | {'key'}
    simple_fields = list(fields - complex_fields)
    regexp = re.sub(r'<({})>'.format('|'.join(simple_fields)),
                    r'(?P<\1>[^/]+)',
                    regexp)
    for field in number_fields:
        if field in fields:
            regexp = re.sub(r'<{}>'.format(field),
                            r'(?P<{}>\d+)'.format(field),
                            regexp)

    if 'key' in fields:
        regexp = re.sub(r'<key>', r'(?P<key>\d+[ABab])', regexp)
    return regexp + '$'


def get_fields(pattern):
    fields = {(field.split('.')[0],
               field.split('.')[1] if len(field.split('.')) == 2 else None)
              for field in re.findall(r'<(.*?)>', pattern)}
    return {field: formatter
            for field, formatter in fields
            if field in FIELDS}
